# Copyright (c) 2018-2024, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from typing import IO, Any, Protocol, Union, runtime_checkable

from ._compat.typing import Traversable

__all__ = (
    "StringOrIO",
    "HasOpen",
    "HasRead",
)


StringOrIO = Union[Traversable, IO[str], str]


@runtime_checkable
class HasOpen(Protocol):
    def open(self) -> Any:
        pass


@runtime_checkable
class HasRead(Protocol):
    def read(self) -> str:
        pass



from __future__ import annotations

import particle


def test_top_level_api():
    assert dir(particle) == [
        "Charge",
        "Corsika7ID",
        "Geant3ID",
        "Inv",
        "InvalidParticle",
        "PDGID",
        "Parity",
        "Particle",
        "ParticleNotFound",
        "PythiaID",
        "SpinType",
        "Status",
        "__version__",
        "latex_to_html_name",
        "lifetime_to_width",
        "width_to_lifetime",
    ]


def test_api_converters():
    assert dir(particle.converters) == [
        "Corsika72PDGIDBiMap",
        "EvtGen2PDGNameMap",
        "EvtGenName2PDGIDBiMap",
        "Geant2PDGIDBiMap",
        "PDG2EvtGenNameMap",
        "Pythia2PDGIDBiMap",
    ]


def test_api_corsika():
    assert dir(particle.corsika) == ["Corsika7ID"]


def test_api_data():
    assert dir(particle.data) == ["basepath"]


def test_api_geant():
    assert dir(particle.geant) == ["Geant3ID"]


def test_api_lhcb():
    assert dir(particle.lhcb) == [
        "LHCbName2PDGIDBiMap",
        "from_lhcb_name",
        "to_lhcb_name",
    ]


def test_api_lhcb_data():
    assert dir(particle.lhcb.data) == ["basepath"]


def test_api_particle():
    assert dir(particle.particle) == [
        "Charge",
        "Inv",
        "InvalidParticle",
        "Parity",
        "Particle",
        "ParticleNotFound",
        "SpinType",
        "Status",
        "latex_name_unicode",
        "latex_to_html_name",
        "lifetime_to_width",
        "programmatic_name",
        "width_to_lifetime",
    ]


def test_api_pdgid():
    assert dir(particle.pdgid) == sorted(
        [
            "PDGID",
            "is_valid",
            "abspid",
            # #
            "is_Qball",
            "is_Rhadron",
            "is_SUSY",
            "is_baryon",
            "is_diquark",
            "is_dyon",
            "is_excited_quark_or_lepton",
            "is_gauge_boson_or_higgs",
            "is_generator_specific",
            "is_hadron",
            "is_lepton",
            "is_meson",
            "is_nucleus",
            "is_pentaquark",
            "is_quark",
            "is_sm_gauge_boson_or_higgs",
            "is_sm_lepton",
            "is_sm_quark",
            "is_special_particle",
            "is_technicolor",
            "has_down",
            "has_up",
            "has_strange",
            "has_charm",
            "has_bottom",
            "has_top",
            "has_fundamental_anti",
            "charge",
            "three_charge",
            "j_spin",
            "J",
            "s_spin",
            "S",
            "l_spin",
            "L",
            "A",
            "Z",
        ]
    )


def test_api_pythia():
    assert dir(particle.pythia) == ["PythiaID"]



# Copyright (c) 2018-2024, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import particle


def test_package_import():
    assert particle is not None


def test_load_version():
    assert particle.__version__



# Copyright (c) 2018-2024, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import pytest

from particle.particle.utilities import (
    greek_letter_name_to_unicode,
    latex_name_unicode,
    str_with_unc,
)

possibilities = (
    (1.234567, 0.01, None, "1.235 ± 0.010"),
    (1.234567e-9, 0.01e-9, None, "1.235e-09 ± 1.0e-11"),
    (1.234567e9, 0.04e9, None, "1.23e+09 ± 4e+07"),
    (0.001, 0.00001, None, "1.000e-03 ± 1.0e-05"),
    (0.00099, 0.00001, None, "9.90e-04 ± 1.0e-05"),
    (99, 0.24, None, "99.00 ± 0.24"),
    (100, 0.25, None, "100.0 ± 0.2"),
    (101, 0.26, None, "101.0 ± 0.3"),
    (0.00001231, 0.000002, 0.000004, "1.23e-05 + 2.0e-06 - 4.0e-06"),
    (1234.5, 0.03, 0.03, "1234.50 ± 0.03"),
    (1234.5, 5, 5, "1234 ± 5"),
    (1234.5, 2, 2, "1234.5 ± 2.0"),
    (1234.5, None, None, "1234.5"),
    (1234.5, None, 2, "1234.5"),
)


@pytest.mark.parametrize(("value", "err_u", "err_l", "test_str"), possibilities)
def test_unc_printout(value, err_u, err_l, test_str):
    assert str_with_unc(value, err_u, err_l) == test_str


possibilities = (
    ("\\omega", "ω"),
    ("\\Omega", "Ω"),
    ("\\Lambda", "Λ"),
    ("\\alpha_{x}^{0}\\beta\\Gamma(1234)\\Omega", "α_{x}^{0}βΓ(1234)Ω"),
)


@pytest.mark.parametrize(("name", "unicode_name"), possibilities)
def test_latex_name_unicode(name, unicode_name):
    assert latex_name_unicode(name) == unicode_name


def test_greek_letter_name_to_unicode():
    """
    Test the one exception that is not verified
    in the test "test_latex_name_unicode" above.
    """
    with pytest.raises(KeyError):
        _ = greek_letter_name_to_unicode("Lambda")
    with pytest.raises(KeyError):
        _ = greek_letter_name_to_unicode("NonExistent")



# Copyright (c) 2018-2024, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import pytest

# Requires pandas
pd = pytest.importorskip("pandas")

from collections import Counter

from particle import data
from particle.particle import Particle
from particle.particle.convert import produce_files
from particle.pdgid import (
    is_diquark,
    is_dyon,
    is_excited_quark_or_lepton,
    is_generator_specific,
    is_lepton,
    is_meson,
    is_pentaquark,
    is_Qball,
    is_quark,
    is_Rhadron,
    is_sm_lepton,
    is_special_particle,
    is_SUSY,
    is_technicolor,
    three_charge,
)

FILES = ["particle2023.csv", "particle2024.csv"]


def test_generate(tmp_path):
    "This verifies that the input and output files match."

    particle2023 = tmp_path / "particle2023.csv"
    particle2024 = tmp_path / "particle2024.csv"

    produce_files(particle2023, particle2024, "DUMMY", "2024")

    particle2024_data = data.basepath / "particle2024.csv"
    with particle2024.open() as src, particle2024_data.open() as res:
        src_filtered = [line for line in src.readlines() if not line.startswith("#")]
        res_filtered = [line for line in res.readlines() if not line.startswith("#")]
        assert src_filtered == res_filtered


@pytest.mark.parametrize("filename", FILES)
def test_csv_file_duplicates(filename):
    particle_data = data.basepath / filename
    p = pd.read_csv(particle_data, comment="#")

    duplicates = {item for item, count in Counter(p.ID).items() if count > 1}
    assert duplicates == set()


@pytest.mark.parametrize("filename", FILES)
def test_csv_file_has_latex(filename):
    particle_data = data.basepath / filename
    p = pd.read_csv(particle_data, comment="#")

    assert p[p.Latex == ""].empty


def test_None_masses():
    "Only certain specific particles should have None masses."
    none_masses = {
        100321,
        -100321,  # K(1460)+
    }
    for p in Particle.all():
        pdgid = p.pdgid
        if pdgid in none_masses:
            continue

        if (
            is_Qball(pdgid)
            or is_Rhadron(pdgid)
            or is_SUSY(pdgid)
            or is_diquark(pdgid)
            or is_dyon(pdgid)
            or is_excited_quark_or_lepton(pdgid)
            or is_generator_specific(pdgid)
            or (is_lepton(pdgid) and three_charge(pdgid) == 0)
            or (is_lepton(pdgid) and not is_sm_lepton(pdgid))  # neutrinos
            or (  # 4-th generation leptons
                is_meson(pdgid) and (abs(pdgid) // 1000000 % 10 == 9)
            )
            or is_pentaquark(  # Mesons with PDGIDs of the kind 9XXXXXX are not experimentally well-known
                pdgid
            )
            or is_quark(pdgid)
            or is_special_particle(pdgid)
            or is_technicolor(pdgid)
        ):
            continue

        assert p.mass is not None


check_nucleons = (
    (2212, 1000010010),
    (-2212, -1000010010),
    (2112, 1000000010),
    (-2112, -1000000010),
)


@pytest.mark.parametrize(("id_particle", "id_nucleus"), check_nucleons)
def test_nucleon_properties(id_particle, id_nucleus):
    """
    Protons and neutrons are both available in the particles table and in the nuclei table
    under IDs 2212 and 2112, and 1000010010 and 1000000010, respectively.
    This trivial test checks most of what is likely to change between PDG updates,
    to ensure consistency.
    """
    p_particle = Particle.from_pdgid(id_particle)
    p_nucleus = Particle.from_pdgid(id_nucleus)

    # Trivial replacement of IDs to avoid obvious irrelevant differences
    assert p_particle.describe().replace(
        f"{id_particle:<12}", f"{id_particle:<12}"
    ) == p_nucleus.describe().replace(f"{id_nucleus:<12}", f"{id_particle:<12}")



# Copyright (c) 2018-2024, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

import sys

if sys.version_info < (3, 9):
    from importlib_resources.abc import Traversable
elif sys.version_info < (3, 11):
    from importlib.abc import Traversable
else:
    from importlib.resources.abc import Traversable


__all__ = ("Traversable",)



from __future__ import annotations

import pytest

from particle.corsika import Corsika7ID
from particle.exceptions import MatchingIDNotFound
from particle.pdgid import PDGID


def test_class_string_representations():
    pid = Corsika7ID(1)
    assert pid == 1
    assert pid.__str__() == "<Corsika7ID: 1>"


def test_class_return_type():
    assert isinstance(Corsika7ID(3), Corsika7ID)


def test_from_pdgid():
    assert Corsika7ID.from_pdgid(-13) == 5

    assert Corsika7ID.from_pdgid(PDGID(-13)) == 5
    assert Corsika7ID.from_pdgid(PDGID(13)) == Corsika7ID(6)


def test_from_pdgid_non_matching():
    with pytest.raises(MatchingIDNotFound):
        _ = Corsika7ID.from_pdgid(55)


def test_to_pdgid():
    cid = Corsika7ID(5)
    assert cid.to_pdgid() == -13
    assert cid.to_pdgid() == PDGID(-13)


def test_to_pdgid_invalid():
    from particle.particle import InvalidParticle  # pylint: disable=C0415

    with pytest.raises(InvalidParticle):
        _ = Corsika7ID(75).to_pdgid()


def test_is_particle():
    cid = Corsika7ID(1)
    assert cid.is_particle()
    cid = Corsika7ID(75)
    assert not cid.is_particle()


def test_from_particle_description():
    cid, is_mother = Corsika7ID.from_particle_description(-6001)
    assert is_mother
    assert cid.is_particle()
    cid, is_mother = Corsika7ID.from_particle_description(75001)
    assert not is_mother
    assert not cid.is_particle()

    # Corsika ID 201 is deuteron
    assert Corsika7ID.from_particle_description(-201000)[0].name() == "D2"


def test_from_particle_description_non_valid():
    with pytest.raises(MatchingIDNotFound):
        _ = Corsika7ID.from_particle_description(0)


def test__is_non_particle_id():
    # Muons
    assert not Corsika7ID._is_non_particle_id(5)
    assert not Corsika7ID._is_non_particle_id(6)
    # Additional muon info, which is not a particle
    assert Corsika7ID._is_non_particle_id(75)
    assert Corsika7ID._is_non_particle_id(76)
    # Weights of the MULTITHIN option
    assert Corsika7ID._is_non_particle_id(8888420)
    # Cherenkov photons on output file
    assert Corsika7ID._is_non_particle_id(9900)


def test_name():
    # check name from pdgid module
    cid = Corsika7ID(5)
    assert cid.name() == "mu+"

    # check name for non-particles
    cid = Corsika7ID(8888420)
    assert cid.name() == "weights of preceding particle (MULTITHIN option)"

    cid = Corsika7ID(9900)
    assert cid.name() == "Cherenkov photons on particle output file"

    cid = Corsika7ID(85)
    assert cid.name() == "decaying μ+ at start"


def test_name_invalid():
    from particle.particle import InvalidParticle  # pylint: disable=C0415

    with pytest.raises(InvalidParticle):
        _ = Corsika7ID(0).name()
